#!/bin/sh

NO_REBOOT=$1
LOCK_FILE=/tmp/lock_multiple_upgrade
CUR_BOOTBANK=
CUR_TRY_BANK_1=
CUR_TRY_BANK_2=
CUR_STAGE=0
BOOTLOADER=0
FIRMWARE=0

DEVICE_BOOT=/dev/mtd0
DEVICE_ENV=/dev/mtd1
BLOCK_SIZE=0x20000
ENV1_OFFSET=0
ENV2_OFFSET=0x40000
ENV_SIZE_IN_BLOCKS=2
FIP1_OFFSET=0x0
FIP2_OFFSET=0x100000
FIP_SIZE_IN_BLOCKS=8

fip_file="fip"
env_file="u-boot-env"
dtb_file="dtb"
kernel_file="kernel"
rootfs_file="rootfs"

VOLUME_NAME__DTB="ubi_dtb"
VOLUME_NAME__KERNEL="ubi_kernel"
VOLUME_NAME__ROOTFS="ubi_rootfs"

firmware_mtd_list="$dtb_file $kernel_file $rootfs_file"
bootloader_mtd_list="$fip_file $env_file"

read_env_variables() {
  echo "Reading variables..." > /dev/console

  CUR_BOOTBANK=$(($(ubootenv -g "cur_bootbank" | cut -d '=' -f2)))
  CUR_TRY_BANK_1=$(($(ubootenv -g "try_bank1" | cut -d '=' -f2)))
  CUR_TRY_BANK_2=$(($(ubootenv -g "try_bank2" | cut -d '=' -f2)))
  
  echo "  cur_bootbank is ${CUR_BOOTBANK}" > /dev/console
  echo "  try_bank1 is ${CUR_TRY_BANK_1}" > /dev/console
  echo "  try_bank2 is ${CUR_TRY_BANK_2}" > /dev/console
}

save_env_variables() {
  echo "Saving variables..." > /dev/console

  ubootenv -s cur_bootbank ${CUR_BOOTBANK}
  ubootenv -s try_bank1 ${CUR_TRY_BANK_1}
  ubootenv -s try_bank2 ${CUR_TRY_BANK_2}
}

check_md5_match() {
  file_base=$1

  md5sum1=$(awk '{ print $1 }' < "./$file_base.md5sum")
  md5sum2=$(md5sum "./$file_base.bin" | awk '{ print $1 }')

  if [ "$md5sum1" != "$md5sum2" ]; then
    echo "[Error] md5sum's isn't equal!!!" > /dev/console
    exit 1
  fi
}

# $1 arg == mtd
# $2 arg == action (0 == write, 1 == erase)
upgrade_failed() {
  CUR_STAGE=$(($CUR_STAGE + 1))
  MTD=$1
  ACTION=$2

  if [[ "$ACTION" == "0" ]]
  then
    echo "  Stage ${CUR_STAGE}: writing ${MTD} has been FAILED" > /dev/console
  else
    echo "  Stage ${CUR_STAGE}: erasing ${MTD} has been FAILED" > /dev/console
  fi

  echo "[Error] Firmware upgrade has been FAILED. Please check files or permissions!" > /dev/console
  exit 1
}

# $1 arg == mtd
# $2 action (0 == write, 1 == erase)
log_upgrade_ok() {
  CUR_STAGE=$(($CUR_STAGE + 1))
  MTD=$1
  ACTION=$2

  if [[ "$ACTION" == "0" ]]
  then
    echo "  Stage ${CUR_STAGE}: writing ${MTD} has been successfully completed" > /dev/console
  else
    echo "  Stage ${CUR_STAGE}: erasing ${MTD} has been successfully completed" > /dev/console
  fi
}

# $1 arg == mtd
# $2 arg == BLOCK_COUNT
erase_flash() {
  MTD=$1
  BLOCK_COUNT=$2

  flash_erase "${MTD}" 0 "${BLOCK_COUNT}" > /dev/null

  if [ "$?" -ne "0" ]; then
    upgrade_failed ${MTD} 1
  else
    log_upgrade_ok ${MTD} 1
  fi
}

# $1 arg == mtd
# $2 arg == path to file
write_flash() {
  MTD=$1
  WRITE_PATH=$2

  nandwrite -pk "${MTD}" "${WRITE_PATH}" > /dev/null

  if [ "$?" -ne "0" ]; then
    upgrade_failed ${MTD} 0
  else
    log_upgrade_ok ${MTD} 0
  fi
}

# $1 arg == device num
# $2 arg == volume name
get_vol_num_from_vol_name() {
	dev_num="$1"
	vol_name="$2"
	info=$(ubinfo -d "${dev_num}" -N "${vol_name}")
	set $info
	echo "/dev/ubi${dev_num}_$3"
}

upgrade_firmware() {
	vdimg=$(get_vol_num_from_vol_name "0" "${VOLUME_NAME__DTB}${CUR_BOOTBANK}")
	vkimg=$(get_vol_num_from_vol_name "0" "${VOLUME_NAME__KERNEL}${CUR_BOOTBANK}")
	vrimg=$(get_vol_num_from_vol_name "0" "${VOLUME_NAME__ROOTFS}${CUR_BOOTBANK}")

	CUR_STAGE=$(($CUR_STAGE + 1))
	echo "  Stage ${CUR_STAGE}: updating $vdimg" > /dev/console
	ubiupdatevol "$vdimg" "$dtb_file.bin"

	CUR_STAGE=$(($CUR_STAGE + 1))
	echo "  Stage ${CUR_STAGE}: updating $vkimg" > /dev/console
	ubiupdatevol "$vkimg" "$kernel_file.bin"

	CUR_STAGE=$(($CUR_STAGE + 1))
	echo "  Stage ${CUR_STAGE}: updating $vrimg" > /dev/console
	ubiupdatevol "$vrimg" "$rootfs_file.bin"
}

upgrade_bootloader() {
  for i in $bootloader_mtd_list; do
    if [ -f "./$i.bin" ]; then
      echo "./$i.bin" > /dev/console

      if [ "$i" == "$fip_file" ]; then
        #TODO: we need to check FIP_SIZE_LIMIT=1048576
        flash_erase ${DEVICE_BOOT} ${FIP1_OFFSET} ${FIP_SIZE_IN_BLOCKS} > /dev/null
        nandwrite -s ${FIP1_OFFSET} -p ${DEVICE_BOOT} "./$i.bin" > /dev/null

        flash_erase ${DEVICE_BOOT} ${FIP2_OFFSET} ${FIP_SIZE_IN_BLOCKS} > /dev/null
        nandwrite -s ${FIP2_OFFSET} -p ${DEVICE_BOOT} "./$i.bin" > /dev/null

        log_upgrade_ok ${DEVICE_BOOT} 0
      fi

      if [ "$i" == "$env_file" ]; then
        flash_erase ${DEVICE_ENV} ${ENV1_OFFSET} ${ENV_SIZE_IN_BLOCKS} > /dev/null
        nandwrite -s ${ENV1_OFFSET} -p ${DEVICE_ENV} "./$i.bin" > /dev/null

        flash_erase ${DEVICE_ENV} ${ENV2_OFFSET} ${ENV_SIZE_IN_BLOCKS} > /dev/null
        nandwrite -s ${ENV2_OFFSET} -p ${DEVICE_ENV} "./$i.bin" > /dev/null

        log_upgrade_ok ${DEVICE_ENV} 0
      fi
    fi
  done
}

upgrade_process() {
  #Step 1. Checking
  for i in $bootloader_mtd_list; do
    if [ -f "./$i.bin" ]; then
      if [ ! -f "./$i.md5sum" ]; then
        echo "[Error] ./$i.md5sum doesn't exist!!!" > /dev/console
        exit 1
      fi
      check_md5_match "$i"

      if [ "$i" == "$fip_file" ] || [ "$i" == "$env_file" ]; then
        BOOTLOADER=1
      fi
    fi
  done

  for i in $firmware_mtd_list; do
    if [ -f "./$i.bin" ]; then
      if [ ! -f "./$i.md5sum" ]; then
        echo "[Error] ./$i.md5sum doesn't exist!!!" > /dev/console
        exit 1
      fi
      check_md5_match "$i"

      if [ "$i" == "$dtb_file" ] || [ "$i" == "$kernel_file" ] || [ "$i" == "$rootfs_file" ]; then
        FIRMWARE=1
      fi
    fi
  done

  if [ "$FIRMWARE" -eq "1" ]; then
    cmp -s version /etc/version

    if [ "$?" == 0 ]; then
      exit 1
    fi
  fi

  #Step 2. Send upgrade notification
#  if [ "$BOOTLOADER" -eq "1" ] || [ "$FIRMWARE" -eq "1" ]; then
#    setupgradestatus 1  #TODO
#  fi

  #Step 3. Start the upgrade process
  if [ "$FIRMWARE" -eq "1" ]; then
    echo "===== Start of FIRMWARE upgrade =====" > /dev/console

    read_env_variables

    # We need to update some uboot environments here
    if [ "$CUR_BOOTBANK" -eq "1" ]; then
      CUR_BOOTBANK=2
    else
      CUR_BOOTBANK=1
    fi

    if [ "$CUR_BOOTBANK" -eq "1" ]; then
      CUR_TRY_BANK_1=0
    else
      CUR_TRY_BANK_2=0
    fi

    upgrade_firmware

    save_env_variables

    echo "===== End of FIRMWARE upgrade =====" > /dev/console
  fi

  if [ "$BOOTLOADER" -eq "1" ]; then
    echo "===== Start of BOOTLOADER upgrade =====" > /dev/console

    read_env_variables

    upgrade_bootloader

    echo "Recovering variables" > /dev/console

    save_env_variables

    echo "===== End of BOOTLOADER upgrade =====" > /dev/console
  fi

  echo "Upgrade has been successfully completed. Well done!" > /dev/console
}

###############################################################################################

if [ -f ${LOCK_FILE} ]; then
  echo "[Warning] You need to reboot to complete the upgrade process!" > /dev/console
  exit 1
fi

upgrade_process

echo "1" > ${LOCK_FILE}

if [[ $NO_REBOOT -ne 1 ]]; then
  echo "The system will be rebooted" > /dev/console
  reboot
fi
